import json

import unittest
from copy import deepcopy

from shinobi_client._common import generate_random_string
from shinobi_client.orms.monitor import ShinobiMonitorOrm, ShinobiMonitorAlreadyExistsError, \
    ShinobiMonitorDoesNotExistError
from shinobi_client.tests._common import TestWithShinobi, _create_email_and_password
from shinobi_client.tests.resources.metadata import get_monitor_configuration

EXAMPLE_MONITOR_1_CONFIGURATION = get_monitor_configuration(1)
EXAMPLE_MONITOR_2_CONFIGURATION = get_monitor_configuration(2)
EXAMPLE_MONITOR_3_CONFIGURATION = get_monitor_configuration(3)


def _create_monitor_id() -> str:
    """
    Creates random monitor identifier.
    :return: created identifier
    """
    return generate_random_string()


class TestShinobiMonitorOrm(TestWithShinobi):
    """
    Tests for `ShinobiMonitorOrm`.
    """
    def setUp(self):
        super().setUp()
        email, password = _create_email_and_password()
        self.user = self.shinobi_client.user.create(email, password)
        self.monitor_orm = ShinobiMonitorOrm(self.superless_shinobi_client, email, password)

    def test_get_when_does_not_exist(self):
        monitor_id = _create_monitor_id()
        self.assertIsNone(self.monitor_orm.get(monitor_id))

    def test_get_all_when_no_monitors(self):
        monitors = self.monitor_orm.get_all()
        self.assertEqual(tuple(), monitors)

    def test_get_all_when_single_monitor(self):
        monitor_id = self._create_monitor()
        monitors = self.monitor_orm.get_all()
        self.assertCountEqual((monitor_id, ), (monitor["mid"] for monitor in monitors))

    def test_get_all_when_multiple_monitors(self):
        monitor_ids = []
        for i in range(3):
            monitor_id = self._create_monitor()
            monitor_ids.append(monitor_id)

        monitors = self.monitor_orm.get_all()
        self.assertCountEqual(monitor_ids, (monitor["mid"] for monitor in monitors))

    def test_create_with_minimal_config(self):
        monitor_id = _create_monitor_id()
        configuration = {"name": EXAMPLE_MONITOR_1_CONFIGURATION["name"], "details": "{}"}
        self.monitor_orm.create(monitor_id, configuration)

        retrieved = self.monitor_orm.get(monitor_id)
        print(retrieved)
        self.assertEqual(EXAMPLE_MONITOR_1_CONFIGURATION["name"], retrieved["name"])
        self.assertEqual("{}", retrieved["details"])

    def test_create(self):
        monitor_id = _create_monitor_id()
        created_monitor = self.monitor_orm.create(monitor_id, EXAMPLE_MONITOR_1_CONFIGURATION)
        self.assertEqual(EXAMPLE_MONITOR_1_CONFIGURATION["host"], created_monitor["host"])
        self.assertEqual(int(EXAMPLE_MONITOR_1_CONFIGURATION["port"]), int(created_monitor["port"]))

        retrieved = self.monitor_orm.get(monitor_id)
        self.assertEqual(monitor_id, retrieved["id"])
        self.assertEqual(monitor_id, retrieved["mid"])
        self.assertEqual(EXAMPLE_MONITOR_1_CONFIGURATION["details"], retrieved["details"])

    def test_create_with_json_details(self):
        monitor_id = _create_monitor_id()
        self.monitor_orm.create(monitor_id, EXAMPLE_MONITOR_3_CONFIGURATION)
        retrieved = self.monitor_orm.get(monitor_id)
        self.assertEqual(EXAMPLE_MONITOR_3_CONFIGURATION["details"], json.loads(retrieved["details"]))

    def test_create_with_multiple_users(self):
        user_1_orm = self.monitor_orm
        user_2 = self.shinobi_client.user.create(*_create_email_and_password())
        user_2_orm = ShinobiMonitorOrm(self.superless_shinobi_client, user_2["email"], user_2["password"])

        monitor_id = _create_monitor_id()
        user_1_orm.create(monitor_id, EXAMPLE_MONITOR_1_CONFIGURATION)

        self.assertEqual(0, len(user_2_orm.get_all()))

    def test_create_with_no_details(self):
        configuration = {"name": "example"}
        self.assertRaises(ValueError, self.monitor_orm.create, _create_monitor_id(), configuration)

    def test_create_with_details_not_json(self):
        configuration = {"name": "example", "details": ""}
        self.assertRaises(ValueError, self.monitor_orm.create, _create_monitor_id(), configuration)

    def test_create_with_details_not_json_object(self):
        configuration = {"name": "example", "details": "false"}
        self.assertRaises(ValueError, self.monitor_orm.create, _create_monitor_id(), configuration)

    def test_create_with_no_name(self):
        configuration = {"details": "{}"}
        self.assertRaises(ValueError, self.monitor_orm.create, _create_monitor_id(), configuration)

    def test_create_when_already_exists(self):
        monitor_id = self._create_monitor()
        self.assertRaises(ShinobiMonitorAlreadyExistsError,
                          self.monitor_orm.create, monitor_id, EXAMPLE_MONITOR_1_CONFIGURATION)

    def test_modify_when_not_exist(self):
        monitor_id = _create_monitor_id()
        self.assertRaises(ShinobiMonitorDoesNotExistError,
                          self.monitor_orm.modify, monitor_id, EXAMPLE_MONITOR_1_CONFIGURATION)

    def test_modify(self):
        for key in EXAMPLE_MONITOR_1_CONFIGURATION.keys():
            with self.subTest(modified=key):
                monitor_id = _create_monitor_id()
                original_monitor = self.monitor_orm.create(monitor_id, EXAMPLE_MONITOR_1_CONFIGURATION)

                new_configuration = deepcopy(EXAMPLE_MONITOR_1_CONFIGURATION)
                new_configuration[key] = EXAMPLE_MONITOR_2_CONFIGURATION[key]
                assert original_monitor[key] != new_configuration[key]

                modified = self.monitor_orm.modify(monitor_id, new_configuration)
                self.assertTrue(modified)

                modified_monitor = self.monitor_orm.get(monitor_id)
                self.assertEqual(new_configuration[key], str(modified_monitor[key]))

                comparable_expected_configuration = filter(
                    lambda item: item[0] != key, ShinobiMonitorOrm.filter_only_supported_keys(
                        new_configuration).items())
                comparable_actual_configuration = map(lambda item: (item[0], str(item[1])), filter(
                    lambda item: item[0] != key, ShinobiMonitorOrm.filter_only_supported_keys(modified_monitor).items()))
                self.assertCountEqual(comparable_expected_configuration, comparable_actual_configuration)

    def test_modify_to_same(self):
        monitor_id = _create_monitor_id()
        self.monitor_orm.create(monitor_id, EXAMPLE_MONITOR_1_CONFIGURATION)
        modified = self.monitor_orm.modify(monitor_id, EXAMPLE_MONITOR_1_CONFIGURATION)
        self.assertFalse(modified)

    def test_modify_to_subset(self):
        monitor_id = _create_monitor_id()
        created = self.monitor_orm.create(monitor_id, EXAMPLE_MONITOR_1_CONFIGURATION)
        modified = self.monitor_orm.modify(
            monitor_id, {"name": EXAMPLE_MONITOR_1_CONFIGURATION["name"], "details": "{}"})
        self.assertFalse(modified)
        self.assertEqual(created, self.monitor_orm.get(monitor_id))

    def test_delete_when_not_exists(self):
        monitor_id = _create_monitor_id()
        self.assertFalse(self.monitor_orm.delete(monitor_id))

    def test_delete(self):
        monitor_id = self._create_monitor()
        self.assertTrue(self.monitor_orm.delete(monitor_id))
        self.assertIsNone(self.monitor_orm.get(monitor_id))

    def test_get_user(self):
        self.assertEqual(self.user["email"], self.monitor_orm.user["email"])

    def _create_monitor(self):
        monitor_id = _create_monitor_id()
        self.monitor_orm.create(monitor_id, EXAMPLE_MONITOR_1_CONFIGURATION)
        return monitor_id


if __name__ == "__main__":
    unittest.main()



import unittest

from shinobi_client.api_key import ShinobiApiKey
from shinobi_client import ShinobiWrongPasswordError
from shinobi_client.tests._common import TestWithShinobi


class TestShinobiApiKey(TestWithShinobi):
    """
    Tests for `ShinobiApiKey`.
    """
    def setUp(self):
        super().setUp()
        self.api_key = ShinobiApiKey(self.superless_shinobi_client)

    def test_get_for_non_existent_user(self):
        self.assertRaises(ShinobiWrongPasswordError, self.api_key.get, "does_not@exist", "password")

    def test_get_with_wrong_password(self):
        user = self._create_user()
        self.assertRaises(ShinobiWrongPasswordError, self.api_key.get, user["email"], "invalid")

    def test_get_api_key(self):
        user = self._create_user()
        api_key = self.api_key.get(user["email"], user["password"])
        self.assertGreater(len(api_key), 0)


if __name__ == "__main__":
    unittest.main()



from shinobi_client.client import ShinobiClient
from shinobi_client.orms.user import ShinobiUserOrm, ShinobiWrongPasswordError
from shinobi_client._common import ShinobiSuperUserCredentialsRequiredError
from shinobi_client.orms.monitor import ShinobiMonitorOrm, ShinobiMonitorAlreadyExistsError
from shinobi_client.api_key import ShinobiApiKey
try:
    from shinobi_client.shinobi_controller import start_shinobi, ShinobiController
except ImportError:
    pass



import string
from asyncio import sleep
from typing import Callable
import random

from requests import Response


class ShinobiSuperUserCredentialsRequiredError(RuntimeError):
    """
    Raised if an operation requires super user credentials that have not been supplied.
    """


def raise_if_errors(shinobi_response: Response, raise_if_json_not_ok: bool = True):
    """
    Raises an exception if the response from Shinobi indicated there were errors.
    :param shinobi_response: the response from Shinobi
    :param raise_if_json_not_ok: raise an exception if Shinobi returns a 2XX but the JSON has `"ok": false`
    """
    shinobi_response.raise_for_status()
    json_response = shinobi_response.json()
    if raise_if_json_not_ok and not json_response["ok"]:
        # Yes, the API returns a 2XX when everything is not ok...
        message = json_response.get("msg", json_response)
        raise RuntimeError(message)


def wait_and_verify(verifier: Callable[[], bool], *, wait_iterations: int = 10,
                    iteration_wait_in_milliseconds_multiplier: int = 100):
    """
    Wait to verify an event has happened.
    :param verifier: callable that returns `True` if the event been waiting upon has been verified
    :param wait_iterations: number of times to try if state not valid
    :param iteration_wait_in_milliseconds_multiplier: waiting for `wait_iterations` * `iteration_wait_in_milliseconds_multiplier`
                                                      between each iteration
    :return: `True` if the event has been verified, else `False`
    """
    for i in range(wait_iterations):
        if verifier():
            return True
        sleep(iteration_wait_in_milliseconds_multiplier * i)
    return False


def generate_random_string(length: int = 8) -> str:
    """
    Generates a short random string.
    :param length: length of generated string
    :return: generated string
    """
    alphabet = string.ascii_lowercase
    return "".join(random.choices(alphabet, k=length))



import json
import os
from typing import Dict, Any

_RESOURCE_DIRECTORY = os.path.dirname(os.path.realpath(__file__))

_MONITORS_DIRECTORY = os.path.join(_RESOURCE_DIRECTORY, "monitors")
EXAMPLE_MONITOR_LOCATIONS = {
    1: os.path.join(_MONITORS_DIRECTORY, "example-1.json"),
    2: os.path.join(_MONITORS_DIRECTORY, "example-2.json"),
    3: os.path.join(_MONITORS_DIRECTORY, "details-as-json.json"),
}


def get_monitor_configuration(example_monitor_id: Any = 1) -> Dict:
    """
    Gets the example monitor configuration with the given ID (or first monitor if no ID is given).
    :param example_monitor_id: ID of monitor to load
    :return: loaded monitor
    """
    return json.load(open(EXAMPLE_MONITOR_LOCATIONS[example_monitor_id], "r"))




from dataclasses import dataclass


@dataclass
class ShinobiClient:
    """
    Shinobi client.

    Not thread safe.
    """
    host: str
    port: str
    super_user_token: str = None
    super_user_email: str = None
    super_user_password: str = None

    @property
    def url(self) -> str:
        return f"http://{self.host}:{self.port}"

    @property
    def user(self) -> "ShinobiUserOrm":
        from shinobi_client.orms.user import ShinobiUserOrm
        return ShinobiUserOrm(self)

    @property
    def api_key(self) -> "ShinobiApiKey":
        from shinobi_client.api_key import ShinobiApiKey
        return ShinobiApiKey(self)

    def monitor(self, email: str, password: str) -> "ShinobiMonitor":
        from shinobi_client.orms.monitor import ShinobiMonitorOrm
        return ShinobiMonitorOrm(self, email, password)



import unittest

from shinobi_client.tests._common import TestWithShinobi, _create_email_and_password


class TestShinobiClient(TestWithShinobi):
    """
    Tests for `ShinobiClient`.
    """
    def test_get_user(self):
        email, password = _create_email_and_password()
        self.shinobi_client.user.create(email, password)
        self.assertIsNotNone(self.shinobi_client.user.get(email))

    def test_get_api_key(self):
        email, password = _create_email_and_password()
        self.shinobi_client.user.create(email, password)
        self.assertIsNotNone(self.shinobi_client.api_key.get(email, password))


if __name__ == "__main__":
    unittest.main()



from typing import Dict

from shinobi_client.client import ShinobiClient


class ShinobiApiKey:
    """
    API key ORM.

    Not thread safe.
    """
    def __init__(self, shinobi_client: ShinobiClient):
        """
        Constructor
        :param shinobi_client: client for Shinobi
        """
        self.shinobi_client = shinobi_client

    def get(self, email: str, password: str) -> Dict:
        """
        Gets the API key for the user with the given email address and password.
        :param email: user's email address
        :param password: user's password
        :return: the user's API key
        :raises ShinobiWrongPasswordError: raised if an incorrect email/password pair is supplied
        """
        return self.shinobi_client.user.get(email, password)["auth_token"]



import unittest
from abc import ABCMeta
from typing import Tuple, ClassVar, Dict

from shinobi_client._common import generate_random_string
from shinobi_client.shinobi_controller import ShinobiController
from shinobi_client.client import ShinobiClient


def _create_email_and_password() -> Tuple[str, str]:
    """
    Create email and password.
    :return: email and password tuple
    """
    random_string = generate_random_string()
    return f"{random_string}@example.com", random_string


class TestWithShinobi(unittest.TestCase, metaclass=ABCMeta):
    """
    Superclass for tests that use Shinobi`.

    Shinobi takes a reasonable amount of time to setup so sharing between all tests in class (at the cost of reduced
    test isolation) might make sense.
    """
    _shinobi_controller_singleton: ClassVar[ShinobiController]
    _shinobi_client_singleton: ClassVar[ShinobiClient]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._shinobi_controller_singleton = ShinobiController()
        cls._shinobi_client_singleton = cls._shinobi_controller_singleton.start()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls._shinobi_controller_singleton.stop()

    def setUp(self):
        super().setUp()
        self.shinobi_client: ShinobiClient = self.__class__._shinobi_client_singleton
        self.superless_shinobi_client = ShinobiClient(self.shinobi_client.host, self.shinobi_client.port)

    def _create_user(self) -> Dict:
        """
        Creates a user.
        :return: model of the created user
        """
        email, password = _create_email_and_password()
        return self.shinobi_client.user.create(email, password)



import json
from copy import deepcopy
from dataclasses import dataclass

from typing import Optional, Dict, Tuple

import requests

from shinobi_client import ShinobiClient
from shinobi_client._common import raise_if_errors, ShinobiSuperUserCredentialsRequiredError, wait_and_verify


@dataclass
class ShinobiWrongPasswordError(ValueError):
    email: str
    password: str


@dataclass
class ShinobiUserAlreadyExistsError(ValueError):
    email: str


@dataclass
class ShinobiUserDoesNotExistError(ValueError):
    email: str


class ShinobiUserOrm:
    """
    Shinobi user ORM.

    Not thread safe.
    """
    @staticmethod
    def _create_improved_user_entry(user: Dict) -> Dict:
        """
        Creates a copy of the given user details with extra information.
        :param user: user details to improve
        :return: improved user details
        """
        user = deepcopy(user)
        assert "email" not in "mail"
        user["email"] = user["mail"]
        if "pass" in user:
            assert "password" not in user
            user["password"] = user["pass"]
        if "ok" in user:
            del user["ok"]
        return user

    @property
    def _base_url(self) -> str:
        if self.shinobi_client.super_user_token is None:
            raise ShinobiSuperUserCredentialsRequiredError()
        return f"http://{self.shinobi_client.host}:{self.shinobi_client.port}" \
               f"/super/{self.shinobi_client.super_user_token}/accounts"

    def __init__(self, shinobi_client: ShinobiClient):
        """
        Constructor.
        :param shinobi_client: client connected to Shinobi installation
        """
        self.shinobi_client = shinobi_client

    def get(self, email: str, password: str = None) -> Optional[Dict]:
        """
        Gets details about the user with the given email address.
        :param email: user's email address
        :param password: password of the user - if supplied will get the details as the user opposed to the super user
        :return: details about user else `None` if the user does not exist
        :raises ShinobiWrongPasswordError: raised if an incorrect email/password pair is supplied
        """
        return self._get_as_user(email, password) if password else self._get_as_super_user(email)

    def _get_as_user(self, email: str, password: str) -> Dict:
        """
        Gets details about the user with the given email address, using the user's own credentials.
        :param email: user's email address
        :param password: user's password
        :return: details about user
        :raises ShinobiWrongPasswordError: raised if an incorrect email/password pair is supplied
        """
        response = requests.post(
            f"http://{self.shinobi_client.host}:{self.shinobi_client.port}/?json=true",
            data={
                "mail": email,
                "pass": password
            })
        raise_if_errors(response, raise_if_json_not_ok=False)

        if not response.json()["ok"]:
            # We can only assume this means that the password was incorrect/the user doesn't exist...
            raise ShinobiWrongPasswordError(email, password)

        user = response.json()["$user"]
        user["pass"] = password
        return ShinobiUserOrm._create_improved_user_entry(user)

    def _get_as_super_user(self, email: str) -> Optional[Dict]:
        """
        Gets details about the user with the given email address, using the super user's credentials.
        :param email: user's email address
        :return: details about user else `None` if the user does not exist
        """
        # XXX: For some reason, Shinobi doesn't have an endpoint to query an individual user as superuser
        users = self.get_all()
        matched_users = tuple(filter(lambda user: user["mail"] == email, users))
        assert len(matched_users) <= 1, f"More than one user found with the email address: {email}"
        if len(matched_users) == 0:
            return None
        return ShinobiUserOrm._create_improved_user_entry(matched_users[0])

    def get_all(self) -> Tuple:
        """
        Gets details about all users.
        :return: tuple where each element contains details about a specific user
        """
        response = requests.get(f"{self._base_url}/list")
        raise_if_errors(response)
        return tuple(ShinobiUserOrm._create_improved_user_entry(user) for user in response.json()["users"])

    def create(self, email: str, password: str, verify: bool = True) -> Dict:
        """
        Creates a user with the given details.
        :param email: email address of the user
        :param password: password for the user
        :param verify: whether to wait to confirm that the user has been created
        :return: details about created user
        """
        # Not trusting Shinobi's API to give back anything useful if the user already exists
        if self.get(email):
            raise ShinobiUserAlreadyExistsError(email)

        # The required post does not align with the API documentation (https://shinobi.video/docs/api)
        # Exploiting the undocumented API successfully used by UI.
        # See source: https://gitlab.com/Shinobi-Systems/Shinobi/-/blob/dev/libs/webServerSuperPaths.js
        data = {
            "mail": email,
            "pass": password,
            "password_again": password,
            "details": json.dumps({
                "factorAuth": "0", "size": "", "days": "", "event_days": "", "log_days": "", "max_camera": "",
                "permissions": "all", "edit_size": "1", "edit_days": "1", "edit_event_days": "1",
                "edit_log_days": "1", "use_admin": "1", "use_aws_s3": "1", "use_whcs": "1", "use_sftp": "1",
                "use_webdav": "1", "use_discordbot": "1", "use_ldap": "1", "aws_use_global": "0",
                "b2_use_global": "0", "webdav_use_global": "0"})
        }
        response = requests.post(f"{self._base_url}/registerAdmin", json=dict(data=data))
        raise_if_errors(response)
        create_user = response.json()

        if verify:
            if not wait_and_verify(lambda: self.get(email) is not None):
                raise RuntimeError("Unable to verify created user")

        return ShinobiUserOrm._create_improved_user_entry(create_user["user"])

    def modify(self, email: str, *, password: str) -> bool:
        """
        Modify a user.
        :param email: the email address of the user to modify
        :param password: new password
        :return: whether the user was modified
        """
        try:
            user = self.get(email, password)
            # The user can't be none as otherwise their credentials shouldn't have worked at getting the user's details
            assert user is not None
            return False
        except ShinobiWrongPasswordError:
            pass

        user = self.get(email)
        if user is None:
            raise ShinobiUserDoesNotExistError(email)

        data = {
            "mail": email,
            "pass": password,
            "password_again": password,
        }
        account = {
            "mail": email,
            "uid": user["uid"],
            "ke": user["ke"]
        }
        response = requests.post(f"{self._base_url}/editAdmin", json=dict(data=data, account=account))
        raise_if_errors(response)

        return True

    def delete(self, email: str, verify: bool = True) -> bool:
        """
        Deletes the user with the given email address.
        :param email: email address of user
        :param verify: whether to wait to confirm that the user has been deleted
        :return: `True` if the user has been deleted, else `False` if they haven't because they didn't exist
        """
        user = self.get(email)
        if user is None:
            return False

        account = {
            "uid": user["uid"],
            "ke": user["ke"],
            "mail": email
        }

        # Odd interface, defined here:
        # https://gitlab.com/Shinobi-Systems/Shinobi/-/blob/dev/libs/webServerSuperPaths.js#L385
        response = requests.post(f"{self._base_url}/deleteAdmin", json=dict(account=account))
        raise_if_errors(response)

        if verify:
            if not wait_and_verify(lambda: self.get(email) is None):
                raise RuntimeError(f"User with email \"{email}\" was not deleted")

        return True



import unittest

from shinobi_client.shinobi_controller import start_shinobi
from shinobi_client.api_key import ShinobiApiKey
from shinobi_client.orms.user import ShinobiUserOrm, ShinobiWrongPasswordError, ShinobiUserAlreadyExistsError, \
    ShinobiUserDoesNotExistError
from shinobi_client.tests._common import _create_email_and_password, TestWithShinobi
from shinobi_client._common import ShinobiSuperUserCredentialsRequiredError


class TestShinobiUserOrm(TestWithShinobi):
    """
    Tests for `ShinobiUserOrm`.
    """
    def setUp(self):
        super().setUp()
        self.user_orm = ShinobiUserOrm(self.shinobi_client)
        self.superless_user_orm = ShinobiUserOrm(self.superless_shinobi_client)
        self.api_key = ShinobiApiKey(self.shinobi_client)

    def test_create(self):
        email, password = _create_email_and_password()
        self.user_orm.create(email, password)
        user = self.user_orm.get(email)
        self.assertEqual(email, user["mail"])
        self.assertIsNotNone(self.api_key.get(email, password))

    def test_create_requires_super_user_credentials(self):
        email, password = _create_email_and_password()
        self.assertRaises(ShinobiSuperUserCredentialsRequiredError, self.superless_user_orm.create, email, password)

    def test_create_when_already_exists(self):
        email, password = _create_email_and_password()
        self.user_orm.create(email, password)
        self.assertRaises(ShinobiUserAlreadyExistsError, self.user_orm.create, email, password)

    def test_get_when_does_not_exist(self):
        user = self.user_orm.get("example@doesnotexist.com")
        self.assertIsNone(user)

    def test_get(self):
        user = self._create_user()
        retrieved_user = self.user_orm.get(user["email"])
        self.assertEqual(user["mail"], retrieved_user["mail"])

    def test_get_using_user_credentials(self):
        user = self._create_user()
        retrieved_user = self.superless_user_orm.get(user["email"], user["password"])
        self.assertEqual(user["mail"], retrieved_user["mail"])

    def test_get_using_invalid_user_credentials(self):
        email, password = _create_email_and_password()
        self.assertRaises(ShinobiWrongPasswordError, self.superless_user_orm.get, email, password)

    def test_get_all_when_no_users(self):
        # Need to start clean installation to ensure no other users
        with start_shinobi() as shinobi_client:
            user_orm = ShinobiUserOrm(shinobi_client)
            self.assertEqual((), user_orm.get_all())

    def test_get_all_when_single_user(self):
        # Need to start clean installation to ensure no other users
        with start_shinobi() as shinobi_client:
            user_orm = ShinobiUserOrm(shinobi_client)
            user = user_orm.create(*_create_email_and_password())
            users = user_orm.get_all()
            self.assertEqual(1, len(users))
            self.assertEqual(user["email"], users[0]["email"])

    def test_get_all_when_multiple_users(self):
        emails = []
        for i in range(3):
            emails.append(self._create_user()["mail"])

        matched_users = tuple(filter(lambda user: user["mail"] in emails, self.user_orm.get_all()))
        self.assertEqual(len(emails), len(matched_users))

    def test_get_all_requires_super_user_credentials(self):
        self.assertRaises(ShinobiSuperUserCredentialsRequiredError, self.superless_user_orm.get_all)

    def test_modify_password_when_user_not_exist(self):
        email, password = _create_email_and_password()
        self.assertRaises(ShinobiUserDoesNotExistError, self.user_orm.modify, email, password=password)

    def test_modify_password(self):
        user = self._create_user()
        _, password = _create_email_and_password()
        assert user["pass"] != password
        changed = self.user_orm.modify(user["email"], password=password)
        self.assertTrue(changed is None or changed is True)
        self.assertRaises(ShinobiWrongPasswordError, self.api_key.get, user["email"], user["pass"])
        self.assertIsNotNone(self.api_key.get(user["email"], password))

    def test_delete_when_user_not_exist(self):
        email, _ = _create_email_and_password()
        self.assertFalse(self.user_orm.delete(email))

    def test_delete(self):
        user = self._create_user()
        self.user_orm.delete(user["email"])
        self.assertIsNone(self.user_orm.get(user["email"]))


if __name__ == "__main__":
    unittest.main()



import json
from copy import deepcopy
from dataclasses import dataclass
from json import JSONDecodeError
from typing import Dict, Optional, Set, Tuple, Union

import requests

from shinobi_client.client import ShinobiClient
from shinobi_client._common import raise_if_errors, wait_and_verify


@dataclass
class ShinobiMonitorAlreadyExistsError(ValueError):
    monitor_id: str


@dataclass
class ShinobiMonitorDoesNotExistError(ValueError):
    monitor_id: str


class InvalidConfigurationError(ValueError):
    def __init__(self, error_message: str, configuration: Dict):
        super().__init__(error_message)
        self.error_message = error_message
        self.configuration = configuration


@dataclass
class ShinobiUnsupportedKeysInConfigurationError(InvalidConfigurationError):
    unsupported_keys: Set[str]


class ShinobiMonitorOrm:
    """
    Shinobi monitor ORM.

    Uses API: https://shinobi.video/docs/api#content-add-edit-or-delete-a-monitor

    Not thread safe.
    """
    SUPPORTED_KEYS = {"name", "details", "type", "ext", "protocol", "host", "path", "port", "fps", "mode", "width",
                      "height"}

    @staticmethod
    def filter_only_supported_keys(configuration: Dict) -> Dict:
        """
        Filter the given configuration to only the key/values pairs that can be set.
        :param configuration: configuration to filter (not modified)
        :return: filtered configuration
        """
        return dict(filter(lambda item: item[0] in ShinobiMonitorOrm.SUPPORTED_KEYS, configuration.items()))

    @staticmethod
    def would_configuration_change(configuration: Dict, existing_configuration: Dict) -> bool:
        """
        Determines whether the existing configuration would change if Shinobi is given the new configuration.
        :param configuration: new configuration
        :param existing_configuration: existing configuration
        :return: `True` if the existing configuration would change
        """
        new_keys = set(configuration.keys()) - set(existing_configuration.keys())
        if len(new_keys) > 0:
            return True
        for key, value in configuration.items():
            if key == "details":
                break
            if str(existing_configuration[key]) != str(value):
                return True

        if "details" not in configuration:
            return False

        # Note: details that are missing with either take default values or will not change
        configuration_details = ShinobiMonitorOrm._parse_details(configuration["details"])
        existing_configuration_details = ShinobiMonitorOrm._parse_details(existing_configuration["details"])
        return ShinobiMonitorOrm.would_configuration_change(configuration_details, existing_configuration_details)

    @staticmethod
    def validate_configuration(configuration: Dict):
        """
        Validates that the configuration only contains supported keys.
        :param configuration: the configuration to validate
        :raises InvalidConfigurationError: if the configuration is deemed invalid
        """
        if "name" not in configuration:
            # Shinobi returns `{'ok': False}` (2XX), with no information if the name is omitted
            raise InvalidConfigurationError("\"name\" is not in the configuration", configuration)
        if "details" not in configuration:
            # Shinobi errors if the "details" field is omitted
            raise InvalidConfigurationError(f"\"details\" is not in the configuration", configuration)
        else:
            # "details" must be JSON
            try:
                ShinobiMonitorOrm._parse_details(configuration["details"])
            except (JSONDecodeError, ValueError) as e:
                raise InvalidConfigurationError(
                    "Configuration \"details\" was invalid: {configuration['details']}", configuration) from e

        unsupported_keys = set(configuration.keys()) - ShinobiMonitorOrm.SUPPORTED_KEYS
        if len(unsupported_keys) > 0:
            raise ShinobiUnsupportedKeysInConfigurationError(unsupported_keys)

    @staticmethod
    def _create_improved_monitor_entry(monitor: Dict) -> Dict:
        """
        Creates a copy of the given monitor details with extra information.
        :param monitor: monitor details to improve
        :return: improved monitor details
        """
        monitor = deepcopy(monitor)
        assert "id" not in monitor
        monitor["id"] = monitor["mid"]
        if "ok" in monitor:
            del monitor["ok"]
        return monitor

    @staticmethod
    def _parse_details(details: Union[Dict, str]) -> Dict:
        """
        Parses the details out of the given details.
        :param details: representation of configuration details
        :return: parsed details (may not be a copy)
        :raises ValueError: if the parsed details are not a dictionary
        """
        # Note: in older versions of Shinobi the details were stored in a JSON dumped string.
        #       Newer config is in JSON.
        parsed_details = json.loads(details) if isinstance(details, str) else details
        if not isinstance(parsed_details, dict):
            raise ValueError(f"Details is not a valid JSON object: {parsed_details}")
        return parsed_details

    @property
    def base_url(self) -> str:
        return f"http://{self.shinobi_client.host}:{self.shinobi_client.port}/{self.api_key}"

    @property
    def user(self) -> Dict:
        return deepcopy(self._user)

    def __init__(self, shinobi_client: ShinobiClient, email: str, password: str):
        """
        Constructor.
        :param shinobi_client: client connected to Shinobi installation
        :param email: email of user to setup monitors for
        :param password: password of user to setup monitors for
        :raises ShinobiWrongPasswordError: if the email and password given is incorrect
        """
        self.shinobi_client = shinobi_client
        self._user = self.shinobi_client.user.get(email, password)
        self.api_key = self._user["auth_token"]
        self.group_key = self._user["ke"]

    def get(self, monitor_id: str) -> Optional[Dict]:
        """
        Gets the monitor with the given ID.
        :param monitor_id: ID of the monitor to get
        :return: details about the monitor else `None` if not found
        """
        response = requests.get(f"{self.base_url}/monitor/{self.group_key}/{monitor_id}")
        response.raise_for_status()
        content = response.json()
        if not content:
            return None
        return ShinobiMonitorOrm._create_improved_monitor_entry(content)

    def get_all(self) -> Tuple[Dict]:
        """
        Gets details about all monitors.
        :return: monitors
        """
        response = requests.get(f"{self.base_url}/monitor/{self.group_key}")
        response.raise_for_status()
        json_response = response.json()

        # Yes, the type of response weirdly changes depending on the number of monitors currently set...
        if isinstance(json_response, Dict):
            if len(json_response) == 0:
                return tuple()
            else:
                return (ShinobiMonitorOrm._create_improved_monitor_entry(json_response), )
        else:
            return tuple(ShinobiMonitorOrm._create_improved_monitor_entry(entry) for entry in json_response)

    def create(self, monitor_id: str,  configuration: Dict, verify: bool = True) -> Dict:
        """
        Creates a monitor with the given ID and configuration.
        :param monitor_id: ID of monitor
        :param configuration: configuration of monitor
        :param verify: wait and verify that the monitor has been created if `True`
        :return: details about the created monitor
        :raises MonitorAlreadyExistsError: raised if a monitor with the given ID already exists
        """
        if "-" in monitor_id:
            # Shinobi silently removes dashes so just making them illegal
            raise ValueError("\"monitor_id\" cannot contain \"-\"")

        configuration = ShinobiMonitorOrm.filter_only_supported_keys(configuration)
        ShinobiMonitorOrm.validate_configuration(configuration)
        if self.get(monitor_id):
            raise ShinobiMonitorAlreadyExistsError(monitor_id)

        self._configure(monitor_id, configuration)

        retrieved_monitor = None

        def retrieve_monitor():
            nonlocal retrieved_monitor
            retrieved_monitor = self.get(monitor_id)
            return retrieved_monitor is not None

        if verify:
            if not wait_and_verify(retrieve_monitor):
                raise RuntimeError(f"Could not create monitor \"{monitor_id}\" with configuration: ${configuration}")
            assert retrieved_monitor is not None
        else:
            retrieved_monitor = self.get(monitor_id)

        return retrieved_monitor

    def modify(self, monitor_id: str, configuration: Dict, verify: bool = True) -> bool:
        """
        Modified a monitor with the given ID with the given configuration.
        :param monitor_id: ID of the monitor
        :param configuration: updated configuration of the monitor
        :param verify: wait and verify that the monitor has been modified if `True`
        :return: `True` if the monitor has been modified
        :raises ShinobiMonitorDoesNotExistError: raised if a monitor with the given ID does not exist
        """
        configuration = ShinobiMonitorOrm.filter_only_supported_keys(configuration)
        ShinobiMonitorOrm.validate_configuration(configuration)
        current_configuration = self.get(monitor_id)
        if not current_configuration:
            raise ShinobiMonitorDoesNotExistError(monitor_id)

        if not ShinobiMonitorOrm.would_configuration_change(configuration, current_configuration):
            return False

        self._configure(monitor_id, configuration)

        if verify and not wait_and_verify(lambda: not ShinobiMonitorOrm.would_configuration_change(
                configuration, self.get(monitor_id))):
            raise RuntimeError(f"Could not change configuration of monitor \"{monitor_id}\". "
                               f"Got {self.get(monitor_id)} expected {configuration}")

        return True

    def delete(self, monitor_id: str, verify: bool = True) -> bool:
        """
        Deletes the monitor with the given ID.

        NoOp if the monitor does not exist.
        :param monitor_id: ID of the monitor
        :param verify: wait and verify that the monitor has been deleted if `True`
        :return: `True` if the monitor was deleted
        """
        # Note: if we don"t do this check, Shinobi errors (and the connection hangs) if asked to remove a non-existent
        #       monitor
        if not self.get(monitor_id):
            return False

        response = requests.post(f"{self.base_url}/configureMonitor/{self.group_key}/{monitor_id}/delete")
        raise_if_errors(response)
        if verify and not wait_and_verify(lambda: self.get(monitor_id) is None):
            raise RuntimeError(f"Could not delete monitor: {monitor_id}")

        return True

    def _configure(self, monitor_id: str, configuration: Dict):
        """
        Configures the monitor with the given ID with the given configuration.

        Will create the monitor if it does not exist.
        :param monitor_id: ID of the monitor
        :param configuration: configuration of the monitor
        """
        # Note: Shinobi used to represent "details" as a JSON dumped string but not needs to be JSON
        configuration["details"] = ShinobiMonitorOrm._parse_details(configuration["details"])
        response = requests.post(f"{self.base_url}/configureMonitor/{self.group_key}/{monitor_id}",
                                 json=dict(data=configuration))
        raise_if_errors(response)
